function incBytes(bT)
  local b = {}
  --for i=1,#bT,1 do
  --  b[i] = bAnd(bT[i] + 1, 0xFF)
  --end
  for k,v in pairs(bT) do
    b[k] = v + 1
  end
  return b
end

function decBytes(bT)
  local b = {}
  --for i=1,#bT,1 do
  --  b[i] = bAnd(bT[i] + 1, 0xFF)
  --end
  for k,v in pairs(bT) do
    b[k] = v - 1
  end
  return b
end

-- read 2 bytes incremented by 1
function readIncWord(addr)
  return byteTableToWord(incBytes(readBytes(addr, 2, true)))
end

function writeDecWord(addr, val)
  writeBytes(addr,decBytes(wordToByteTable(val)))
end

function ScanButtonClick(sender)

  base = AOBScanUnique(
    "48 4D 53 44 4B 4B 48 46 44 4D 53 1F 52 58 52 53 " ..
    "44 4C 52 FF FE FE FE FE FE FE FE FE FE FE FE FE " ..
    "4C 60 68 6D 1F 4F 71 6E 66 71 60 6C FF"
  )

  if (base == nil) or (base == 0) then
    Patcher.Status.Caption = "Scanning Failed! :("
    return
  end

  local bIS = readBytes(base, 19, true)
  local bMain = readBytes(base+0x20, 12, true)
  local bVersion = readBytes(base+0x40, 4, true)

  local sIS = byteTableToString(incBytes(bIS))
  local sMain = byteTableToString(incBytes(bMain))
  local nVersion = byteTableToDword(incBytes(bVersion))

  Patcher.Signature.Caption = sIS
  Patcher.Type.Caption = sMain
  Patcher.Version.Caption = string.format("0x%08X",nVersion)
  Patcher.VersionEdit.Text = string.format("0x%X",nVersion+1)

  -- locations of instructions where offsets are stored
  --   they're in the form "movea val, r0, rn", val is stored in the
  --   second 2 bytes of the 4 byte instruction
  -- AV1X: 0x25DE, 0x25E2 (Jrev, Normal)
  -- AV1Y: 0x2650, 0x2654
  -- AV1XDual: 0x2618, 0x261C
  -- AV1YDual: 0x268E
  AV1X = readIncWord(base+0x25E4)
  AV1Y = readIncWord(base+0x2656)
  AV1XDual = readIncWord(base+0x261E)
  AV1YDual = readIncWord(base+0x2690)
  -- AV1Gap: 0x26B0, mov 8 r13
  AV1Gap = readBytes(base+0x26B0, 1) + 1

  -- AV2X: 0x26EA, 0x26EE
  -- AV2Y: 0x2756, 0x275A
  -- AV2XDual: 0x2720, 0x2724
  -- AV2YDual: 0x2790
  AV2X = readIncWord(base+0x26F0)
  AV2Y = readIncWord(base+0x275C)
  AV2XDual = readIncWord(base+0x2726)
  AV2YDual = readIncWord(base+0x2792)
  -- AV2Gap: 0x27B6, mov 8 r13
  AV2Gap = readBytes(base+0x27B6, 1) + 1


  Patcher.AV1XEdit.Text = AV1X
  Patcher.AV1YEdit.Text = AV1Y
  Patcher.AV1XDualEdit.Text = AV1XDual
  Patcher.AV1YDualEdit.Text = AV1YDual
  Patcher.AV1GapEdit.Text = AV1Gap

  Patcher.AV2XEdit.Text = AV2X
  Patcher.AV2YEdit.Text = AV2Y
  Patcher.AV2XDualEdit.Text = AV2XDual
  Patcher.AV2YDualEdit.Text = AV2YDual
  Patcher.AV2GapEdit.Text = AV2Gap

  -- Enable user to make changes
  Patcher.PatchVersionCheck.Enabled = true
  -- Patcher.VersionEdit.Enabled = true
  Patcher.AspectFixPanel.Enabled = true
  Patcher.OffsetPanel.Enabled = true
  Patcher.ApplyButton.Enabled = true

  Patcher.Status.Caption = "You can start selecting patches!"
end

function PatchVersionCheckChange(sender)
  Patcher.VersionEdit.Enabled = Patcher.PatchVersionCheck.Checked
end

function AspectFixCheckChange(sender)
  local state = Patcher.AspectFixCheck.Checked
  Patcher.DIP2Off.Enabled = state
  Patcher.DIP2On.Enabled = state
  Patcher.DIP2Both.Enabled = state
end

function AdjustOffsetCheckChange(sender)
  local state = Patcher.AdjustOffsetCheck.Checked
  Patcher.AV1.Enabled = state
  Patcher.AV2.Enabled = state
  Patcher.OffsetResetButton.Enabled = state
end


function OffsetResetButtonClick(sender)
  Patcher.AV1XEdit.Text = AV1X
  Patcher.AV1YEdit.Text = AV1Y
  Patcher.AV1XDualEdit.Text = AV1XDual
  Patcher.AV1YDualEdit.Text = AV1YDual
  Patcher.AV1GapEdit.Text = AV1Gap

  Patcher.AV2XEdit.Text = AV2X
  Patcher.AV2YEdit.Text = AV2Y
  Patcher.AV2XDualEdit.Text = AV2XDual
  Patcher.AV2YDualEdit.Text = AV2YDual
  Patcher.AV2GapEdit.Text = AV2Gap
end

function ApplyButtonClick(sender)
  local b = false

  if Patcher.PatchVersionCheck.Checked then
    b = true

    writeBytes(base+0x40,decBytes(dwordToByteTable(Patcher.VersionEdit.Text)))
  end

  if Patcher.AspectFixCheck.Checked then
    b = true

    if Patcher.DIP2Off.Checked then
      -- 128CE bnh     loc_12900
      writeBytes(base+0x28CE, 0x94)
    elseif Patcher.DIP2On.Checked then
      -- 1379E bnh     loc_137D0
      writeBytes(base+0x379E, 0x94)
    else -- both
      writeBytes(base+0x28CE, 0x94)
      writeBytes(base+0x379E, 0x94)
    end
  end

  if Patcher.AdjustOffsetCheck.Checked then
    b = true
    -- AV1X: 0x25DE, 0x25E2
    writeDecWord(base+0x25E4, Patcher.AV1XEdit.Text)
    writeDecWord(base+0x25F0, Patcher.AV1XEdit.Text)
    -- AV1Y: 0x2650, 0x2654
    writeDecWord(base+0x2656, Patcher.AV1YEdit.Text)
    writeDecWord(base+0x2652, Patcher.AV1YEdit.Text)
    -- AV1XDual: 0x2618, 0x261C
    writeDecWord(base+0x261E, Patcher.AV1XDualEdit.Text)
    writeDecWord(base+0x261A, Patcher.AV1XDualEdit.Text)
    -- AV1YDual: 0x268E
    writeDecWord(base+0x2690, Patcher.AV1YDualEdit.Text)

    -- AV2X: 0x26EA, 0x26EE
    writeDecWord(base+0x25F0, Patcher.AV2XEdit.Text)
    writeDecWord(base+0x25EC, Patcher.AV2XEdit.Text)
    -- AV2Y: 0x2756, 0x275A
    writeDecWord(base+0x275C, Patcher.AV2YEdit.Text)
    writeDecWord(base+0x2758, Patcher.AV2YEdit.Text)
    -- AV2XDual: 0x2720, 0x2724
    writeDecWord(base+0x2726, Patcher.AV2XDualEdit.Text)
    writeDecWord(base+0x2722, Patcher.AV2XDualEdit.Text)
    -- AV2YDual: 0x2790
    writeDecWord(base+0x2792, Patcher.AV2YDualEdit.Text)

    -- AV1Gap: 0x26B0, mov 8 r13
    writeBytes(base+0x26B0, tonumber(Patcher.AV1GapEdit.Text)-1)
    -- AV2Gap: 0x27B6, mov 8 r13
    writeBytes(base+0x27B6, tonumber(Patcher.AV2GapEdit.Text)-1)
  end

  if b then
    Patcher.Status.Caption = "Patches are applied! Scan to verify offsets!"
  else
    Patcher.Status.Caption = "You haven't selected any patches to apply :/"
  end
end

function ShowAsHexClick(sender)
  Patcher.AV1XEdit.Text = string.format("0x%X", Patcher.AV1XEdit.Text)
  Patcher.AV1YEdit.Text = string.format("0x%X", Patcher.AV1YEdit.Text)
  Patcher.AV1XDualEdit.Text = string.format("0x%X", Patcher.AV1XDualEdit.Text)
  Patcher.AV1YDualEdit.Text = string.format("0x%X", Patcher.AV1YDualEdit.Text)

  Patcher.AV2XEdit.Text = string.format("0x%X", Patcher.AV2XEdit.Text)
  Patcher.AV2YEdit.Text = string.format("0x%X", Patcher.AV2YEdit.Text)
  Patcher.AV2XDualEdit.Text = string.format("0x%X", Patcher.AV2XDualEdit.Text)
  Patcher.AV2YDualEdit.Text = string.format("0x%X", Patcher.AV2YDualEdit.Text)

  Patcher.AV1GapEdit.Text = string.format("0x%X", Patcher.AV1GapEdit.Text);
  Patcher.AV2GapEdit.Text = string.format("0x%X", Patcher.AV2GapEdit.Text);
end

function ShowAsDecimalClick(sender)
  Patcher.AV1XEdit.Text = string.format("%d", Patcher.AV1XEdit.Text)
  Patcher.AV1YEdit.Text = string.format("%d", Patcher.AV1YEdit.Text)
  Patcher.AV1XDualEdit.Text = string.format("%d", Patcher.AV1XDualEdit.Text)
  Patcher.AV1YDualEdit.Text = string.format("%d", Patcher.AV1YDualEdit.Text)

  Patcher.AV2XEdit.Text = string.format("%d", Patcher.AV2XEdit.Text)
  Patcher.AV2YEdit.Text = string.format("%d", Patcher.AV2YEdit.Text)
  Patcher.AV2XDualEdit.Text = string.format("%d", Patcher.AV2XDualEdit.Text)
  Patcher.AV2YDualEdit.Text = string.format("%d", Patcher.AV2YDualEdit.Text)

  Patcher.AV1GapEdit.Text = string.format("%d", Patcher.AV1GapEdit.Text);
  Patcher.AV2GapEdit.Text = string.format("%d", Patcher.AV2GapEdit.Text);
end

function FormClose(sender)
  ----
  ---- comment this out when editing the form
  closeCE()
  return caFree --Possible options: caHide, caFree, caMinimize, caNone
  ----
  ---- uncomment this when editing the form
  --return caHide --Possible options: caHide, caFree, caMinimize, caNone
end

function clamp(low, high, n)
  local x = tonumber(n)

  if (x == nil) then
    x = 0
  elseif (x > high) then
    x = high
  elseif (x < low) then
    x = 0
  else
    x = x
  end

  return x
end

function AV1XEditExit(sender)
  Patcher.AV1XEdit.Text = clamp(0, 0xFF, Patcher.AV1XEdit.Text)
end
function AV1YEditExit(sender)
  Patcher.AV1YEdit.Text = clamp(0, 0xFF, Patcher.AV1YEdit.Text)
end
function AV1XDualEditExit(sender)
  Patcher.AV1XDualEdit.Text = clamp(0, 0xFFFF, Patcher.AV1XDualEdit.Text)
end
function AV1YDualEditExit(sender)
  Patcher.AV1YDualEdit.Text = clamp(0, 0xFF, Patcher.AV1YDualEdit.Text)
end
function AV1GapEditExit(sender)
  Patcher.AV1GapEdit.Text = clamp(0, 31, Patcher.AV1GapEdit.Text)
end

function AV2XEditExit(sender)
  Patcher.AV2XEdit.Text = clamp(0, 0xFF, Patcher.AV2XEdit.Text)
end
function AV2YEditExit(sender)
  Patcher.AV2YEdit.Text = clamp(0, 0xFF, Patcher.AV2YEdit.Text)
end
function AV2XDualEditExit(sender)
  Patcher.AV2XDualEdit.Text = clamp(0, 0xFFFF, Patcher.AV2XDualEdit.Text)
end
function AV2YDualEditExit(sender)
  Patcher.AV2YDualEdit.Text = clamp(0, 0xFF, Patcher.AV2YDualEdit.Text)
end
function AV2GapEditExit(sender)
  Patcher.AV2GapEdit.Text = clamp(0, 31, Patcher.AV2GapEdit.Text)
end

form_show(Patcher)
strings_add(getAutoAttachList(), "ISNP.exe")
